@@ -2,17 +2,22 @@
  * main.c :  Main control function for svnserve
  *
  * ====================================================================
- * Copyright (c) 2000-2007 CollabNet.  All rights reserved.
+ *    Licensed to the Apache Software Foundation (ASF) under one
+ *    or more contributor license agreements.  See the NOTICE file
+ *    distributed with this work for additional information
+ *    regarding copyright ownership.  The ASF licenses this file
+ *    to you under the Apache License, Version 2.0 (the
+ *    "License"); you may not use this file except in compliance
+ *    with the License.  You may obtain a copy of the License at
  *
- * This software is licensed as described in the file COPYING, which
- * you should have received as part of this distribution.  The terms
- * are also available at http://subversion.tigris.org/license-1.html.
- * If newer versions of this license are posted there, you may use a
- * newer version instead, at your option.
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
- * This software consists of voluntary contributions made by many
- * individuals.  For exact contribution history, see the revision
- * history and logs, available at http://subversion.tigris.org/.
+ *    Unless required by applicable law or agreed to in writing,
+ *    software distributed under the License is distributed on an
+ *    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ *    KIND, either express or implied.  See the License for the
+ *    specific language governing permissions and limitations
+ *    under the License.
  * ====================================================================
  */
 
@@ -35,14 +40,17 @@
 #include "svn_error.h"
 #include "svn_ra_svn.h"
 #include "svn_utf.h"
+#include "svn_dirent_uri.h"
 #include "svn_path.h"
 #include "svn_opt.h"
 #include "svn_repos.h"
-#include "svn_fs.h"
+#include "svn_string.h"
+#include "svn_cache_config.h"
 #include "svn_version.h"
 #include "svn_io.h"
 
 #include "svn_private_config.h"
+#include "private/svn_dep_compat.h"
 #include "winservice.h"
 
 #ifdef HAVE_UNISTD_H
@@ -128,46 +136,122 @@ void winservice_notify_stop(void)
  * APR requires that options without abbreviations
  * have codes greater than 255.
  */
-#define SVNSERVE_OPT_LISTEN_PORT 256
-#define SVNSERVE_OPT_LISTEN_HOST 257
-#define SVNSERVE_OPT_FOREGROUND  258
-#define SVNSERVE_OPT_TUNNEL_USER 259
-#define SVNSERVE_OPT_VERSION     260
-#define SVNSERVE_OPT_PID_FILE    261
-#define SVNSERVE_OPT_SERVICE     262
-#define SVNSERVE_OPT_CONFIG_FILE 263
+#define SVNSERVE_OPT_LISTEN_PORT     256
+#define SVNSERVE_OPT_LISTEN_HOST     257
+#define SVNSERVE_OPT_FOREGROUND      258
+#define SVNSERVE_OPT_TUNNEL_USER     259
+#define SVNSERVE_OPT_VERSION         260
+#define SVNSERVE_OPT_PID_FILE        261
+#define SVNSERVE_OPT_SERVICE         262
+#define SVNSERVE_OPT_CONFIG_FILE     263
+#define SVNSERVE_OPT_LOG_FILE        264
+#define SVNSERVE_OPT_CACHE_TXDELTAS  265
+#define SVNSERVE_OPT_CACHE_FULLTEXTS 266
 
 static const apr_getopt_option_t svnserve__options[] =
   {
     {"daemon",           'd', 0, N_("daemon mode")},
-    {"listen-port",       SVNSERVE_OPT_LISTEN_PORT, 1,
-     N_("listen port (for daemon mode)")},
-    {"listen-host",       SVNSERVE_OPT_LISTEN_HOST, 1,
-     N_("listen hostname or IP address (for daemon mode)")},
-    {"foreground",        SVNSERVE_OPT_FOREGROUND, 0,
-     N_("run in foreground (useful for debugging)")},
-    {"help",             'h', 0, N_("display this help")},
-    {"version",           SVNSERVE_OPT_VERSION, 0,
-     N_("show program version information")},
     {"inetd",            'i', 0, N_("inetd mode")},
+    {"tunnel",           't', 0, N_("tunnel mode")},
+    {"listen-once",      'X', 0, N_("listen-once mode (useful for debugging)")},
+#ifdef WIN32
+    {"service",          SVNSERVE_OPT_SERVICE, 0,
+     N_("Windows service mode (Service Control Manager)")},
+#endif
     {"root",             'r', 1, N_("root of directory to serve")},
     {"read-only",        'R', 0,
      N_("force read only, overriding repository config file")},
-    {"tunnel",           't', 0, N_("tunnel mode")},
-    {"tunnel-user",      SVNSERVE_OPT_TUNNEL_USER, 1,
-     N_("tunnel username (default is current uid's name)")},
-#ifdef CONNECTION_HAVE_THREAD_OPTION
-    {"threads",          'T', 0, N_("use threads instead of fork")},
-#endif
-    {"listen-once",      'X', 0, N_("listen once (useful for debugging)")},
     {"config-file",      SVNSERVE_OPT_CONFIG_FILE, 1,
      N_("read configuration from file ARG")},
-    {"pid-file",         SVNSERVE_OPT_PID_FILE, 1,
-     N_("write server process ID to file ARG")},
+    {"listen-port",       SVNSERVE_OPT_LISTEN_PORT, 1,
 #ifdef WIN32
-    {"service",          SVNSERVE_OPT_SERVICE, 0,
-     N_("run as a windows service (SCM only)")},
+     N_("listen port\n"
+        "                             "
+        "[mode: daemon, service, listen-once]")},
+#else
+     N_("listen port\n"
+        "                             "
+        "[mode: daemon, listen-once]")},
 #endif
+    {"listen-host",       SVNSERVE_OPT_LISTEN_HOST, 1,
+#ifdef WIN32
+     N_("listen hostname or IP address\n"
+        "                             "
+        "[mode: daemon, service, listen-once]")},
+#else
+     N_("listen hostname or IP address\n"
+        "                             "
+        "[mode: daemon, listen-once]")},
+#endif
+    {"prefer-ipv6",      '6', 0,
+     N_("prefer IPv6 when resolving the listen hostname\n"
+        "                             "
+        "[IPv4 is preferred by default. Using IPv4 and IPv6\n"
+        "                             "
+        "at the same time is not supported in daemon mode.\n"
+        "                             "
+        "Use inetd mode or tunnel mode if you need this.]")},
+    {"compression",      'c', 1,
+     N_("compression level to use for network transmissions\n"
+        "                             "
+        "[0 .. no compression, 5 .. default, \n"
+        "                             "
+        " 9 .. maximum compression]")},
+    {"memory-cache-size", 'M', 1,
+     N_("size of the extra in-memory cache in MB used to\n"
+        "                             "
+        "minimize redundant operations.\n"
+        "                             "
+        "Default is 128 for threaded and 16 for non-\n"
+        "                             "
+        "threaded mode.\n"
+        "                             "
+        "[used for FSFS repositories only]")},
+    {"cache-txdeltas", SVNSERVE_OPT_CACHE_TXDELTAS, 1,
+     N_("enable or disable caching of deltas between older\n"
+        "                             "
+        "revisions.\n"
+        "                             "
+        "Default is no.\n"
+        "                             "
+        "[used for FSFS repositories only]")},
+    {"cache-fulltexts", SVNSERVE_OPT_CACHE_FULLTEXTS, 1,
+     N_("enable or disable caching of file contents\n"
+        "                             "
+        "Default is yes.\n"
+        "                             "
+        "[used for FSFS repositories only]")},
+#ifdef CONNECTION_HAVE_THREAD_OPTION
+    /* ### Making the assumption here that WIN32 never has fork and so
+     * ### this option never exists when --service exists. */
+    {"threads",          'T', 0, N_("use threads instead of fork "
+                                    "[mode: daemon]")},
+#endif
+    {"foreground",        SVNSERVE_OPT_FOREGROUND, 0,
+     N_("run in foreground (useful for debugging)\n"
+        "                             "
+        "[mode: daemon]")},
+    {"log-file",         SVNSERVE_OPT_LOG_FILE, 1,
+     N_("svnserve log file")},
+    {"pid-file",         SVNSERVE_OPT_PID_FILE, 1,
+#ifdef WIN32
+     N_("write server process ID to file ARG\n"
+        "                             "
+        "[mode: daemon, listen-once, service]")},
+#else
+     N_("write server process ID to file ARG\n"
+        "                             "
+        "[mode: daemon, listen-once]")},
+#endif
+    {"tunnel-user",      SVNSERVE_OPT_TUNNEL_USER, 1,
+     N_("tunnel username (default is current uid's name)\n"
+        "                             "
+        "[mode: tunnel]")},
+    {"help",             'h', 0, N_("display this help")},
+    {"version",           SVNSERVE_OPT_VERSION, 0,
+     N_("show program version information")},
+    {"quiet",            'q', 0,
+     N_("no progress (only errors) to stderr")},
     {0,                  0,   0, 0}
   };
 
@@ -187,10 +271,19 @@ static void help(apr_pool_t *pool)
 {
   apr_size_t i;
 
-  svn_error_clear(svn_cmdline_fputs(_("usage: svnserve [options]\n"
+#ifdef WIN32
+  svn_error_clear(svn_cmdline_fputs(_("usage: svnserve [-d | -i | -t | -X "
+                                      "| --service] [options]\n"
                                       "\n"
                                       "Valid options:\n"),
                                     stdout, pool));
+#else
+  svn_error_clear(svn_cmdline_fputs(_("usage: svnserve [-d | -i | -t | -X] "
+                                      "[options]\n"
+                                      "\n"
+                                      "Valid options:\n"),
+                                    stdout, pool));
+#endif
   for (i = 0; svnserve__options[i].name && svnserve__options[i].optch; i++)
     {
       const char *optstr;
@@ -201,7 +294,7 @@ static void help(apr_pool_t *pool)
   exit(0);
 }
 
-static svn_error_t * version(apr_pool_t *pool)
+static svn_error_t * version(svn_boolean_t quiet, apr_pool_t *pool)
 {
   const char *fs_desc_start
     = _("The following repository back-end (FS) modules are available:\n\n");
@@ -216,8 +309,8 @@ static svn_error_t * version(apr_pool_t *pool)
                            _("\nCyrus SASL authentication is available.\n"));
 #endif
 
-  return svn_opt_print_help(NULL, "svnserve", TRUE, FALSE, version_footer->data,
-                            NULL, NULL, NULL, NULL, pool);
+  return svn_opt_print_help3(NULL, "svnserve", TRUE, quiet, version_footer->data,
+                             NULL, NULL, NULL, NULL, NULL, pool);
 }
 
 
@@ -228,7 +321,9 @@ static void sigchld_handler(int signo)
 }
 #endif
 
-/* In tunnel or inetd mode, we don't want hook scripts corrupting the
+/* Redirect stdout to stderr.  ARG is the pool.
+ *
+ * In tunnel or inetd mode, we don't want hook scripts corrupting the
  * data stream by sending data to stdout, so we need to redirect
  * stdout somewhere else.  Sending it to stderr is acceptable; sending
  * it to /dev/null is another option, but apr doesn't provide a way to
@@ -332,9 +427,14 @@ int main(int argc, const char *argv[])
   apr_uint16_t port = SVN_RA_SVN_PORT;
   const char *host = NULL;
   int family = APR_INET;
+  apr_int32_t sockaddr_info_flags = 0;
+  svn_boolean_t prefer_v6 = FALSE;
+  svn_boolean_t quiet = FALSE;
+  svn_boolean_t is_version = FALSE;
   int mode_opt_count = 0;
   const char *config_filename = NULL;
   const char *pid_filename = NULL;
+  const char *log_filename = NULL;
   svn_node_kind_t kind;
 
   /* Initialize the app. */
@@ -369,6 +469,12 @@ int main(int argc, const char *argv[])
   params.cfg = NULL;
   params.pwdb = NULL;
   params.authzdb = NULL;
+  params.compression_level = SVN_DELTA_COMPRESSION_LEVEL_DEFAULT;
+  params.log_file = NULL;
+  params.username_case = CASE_ASIS;
+  params.memory_cache_size = (apr_uint64_t)-1;
+  params.cache_fulltexts = TRUE;
+  params.cache_txdeltas = FALSE;
 
   while (1)
     {
@@ -379,13 +485,20 @@ int main(int argc, const char *argv[])
         usage(argv[0], pool);
       switch (opt)
         {
+        case '6':
+          prefer_v6 = TRUE;
+          break;
+
         case 'h':
           help(pool);
           break;
 
+        case 'q':
+          quiet = TRUE;
+          break;
+
         case SVNSERVE_OPT_VERSION:
-          SVN_INT_ERR(version(pool));
-          exit(0);
+          is_version = TRUE;
           break;
 
         case 'd':
@@ -409,7 +522,17 @@ int main(int argc, const char *argv[])
           break;
 
         case SVNSERVE_OPT_LISTEN_PORT:
-          port = atoi(arg);
+          {
+            apr_uint64_t val;
+
+            err = svn_cstring_strtoui64(&val, arg, 0, APR_UINT16_MAX, 10);
+            if (err)
+              return svn_cmdline_handle_exit_error(
+                       svn_error_createf(SVN_ERR_CL_ARG_PARSING_ERROR, err,
+                                         _("Invalid port '%s'"), arg),
+                       pool, "svnserve: ");
+            port = (apr_uint16_t)val;
+          }
           break;
 
         case SVNSERVE_OPT_LISTEN_HOST:
@@ -452,8 +575,8 @@ int main(int argc, const char *argv[])
               return EXIT_FAILURE;
             }
 
-          params.root = svn_path_internal_style(params.root, pool);
-          SVN_INT_ERR(svn_path_get_absolute(&params.root, params.root, pool));
+          params.root = svn_dirent_internal_style(params.root, pool);
+          SVN_INT_ERR(svn_dirent_get_absolute(&params.root, params.root, pool));
           break;
 
         case 'R':
@@ -464,6 +587,28 @@ int main(int argc, const char *argv[])
           handling_mode = connection_mode_thread;
           break;
 
+        case 'c':
+          params.compression_level = atoi(arg);
+          if (params.compression_level < SVN_DELTA_COMPRESSION_LEVEL_NONE)
+            params.compression_level = SVN_DELTA_COMPRESSION_LEVEL_NONE;
+          if (params.compression_level > SVN_DELTA_COMPRESSION_LEVEL_MAX)
+            params.compression_level = SVN_DELTA_COMPRESSION_LEVEL_MAX;
+          break;
+
+        case 'M':
+          params.memory_cache_size = 0x100000 * apr_strtoi64(arg, NULL, 0);
+          break;
+
+        case SVNSERVE_OPT_CACHE_TXDELTAS:
+          params.cache_txdeltas
+             = svn_tristate__from_word(arg) == svn_tristate_true;
+          break;
+
+        case SVNSERVE_OPT_CACHE_FULLTEXTS:
+          params.cache_fulltexts
+             = svn_tristate__from_word(arg) == svn_tristate_true;
+          break;
+
 #ifdef WIN32
         case SVNSERVE_OPT_SERVICE:
           if (run_mode != run_mode_service)
@@ -476,27 +621,46 @@ int main(int argc, const char *argv[])
 
         case SVNSERVE_OPT_CONFIG_FILE:
           SVN_INT_ERR(svn_utf_cstring_to_utf8(&config_filename, arg, pool));
-          config_filename = svn_path_internal_style(config_filename, pool);
-          SVN_INT_ERR(svn_path_get_absolute(&config_filename, config_filename,
-                                            pool));
+          config_filename = svn_dirent_internal_style(config_filename, pool);
+          SVN_INT_ERR(svn_dirent_get_absolute(&config_filename, config_filename,
+                                              pool));
           break;
 
         case SVNSERVE_OPT_PID_FILE:
           SVN_INT_ERR(svn_utf_cstring_to_utf8(&pid_filename, arg, pool));
-          pid_filename = svn_path_internal_style(pid_filename, pool);
-          SVN_INT_ERR(svn_path_get_absolute(&pid_filename, pid_filename,
-                                            pool));
+          pid_filename = svn_dirent_internal_style(pid_filename, pool);
+          SVN_INT_ERR(svn_dirent_get_absolute(&pid_filename, pid_filename,
+                                              pool));
+          break;
+
+        case SVNSERVE_OPT_LOG_FILE:
+          SVN_INT_ERR(svn_utf_cstring_to_utf8(&log_filename, arg, pool));
+          log_filename = svn_dirent_internal_style(log_filename, pool);
+          SVN_INT_ERR(svn_dirent_get_absolute(&log_filename, log_filename,
+                                              pool));
           break;
 
         }
     }
+
+  if (is_version)
+    {
+      SVN_INT_ERR(version(quiet, pool));
+      exit(0);
+    }
+
   if (os->ind != argc)
     usage(argv[0], pool);
 
   if (mode_opt_count != 1)
     {
-      svn_error_clear(svn_cmdline_fputs
-                      (_("You must specify exactly one of -d, -i, -t or -X.\n"),
+      svn_error_clear(svn_cmdline_fputs(
+#ifdef WIN32
+                      _("You must specify exactly one of -d, -i, -t, "
+                        "--service or -X.\n"),
+#else
+                      _("You must specify exactly one of -d, -i, -t or -X.\n"),
+#endif
                        stderr, pool));
       usage(argv[0], pool);
     }
@@ -504,10 +668,16 @@ int main(int argc, const char *argv[])
   /* If a configuration file is specified, load it and any referenced
    * password and authorization files. */
   if (config_filename)
-      SVN_INT_ERR(load_configs(&params.cfg, &params.pwdb, &params.authzdb,
-                               config_filename, TRUE,
-                               svn_path_dirname(config_filename, pool),
-                               pool));
+    SVN_INT_ERR(load_configs(&params.cfg, &params.pwdb, &params.authzdb,
+                             &params.username_case, config_filename, TRUE,
+                             svn_dirent_dirname(config_filename, pool),
+                             NULL, NULL, /* server baton, conn */
+                             pool));
+
+  if (log_filename)
+    SVN_INT_ERR(svn_io_file_open(&params.log_file, log_filename,
+                                 APR_WRITE | APR_CREATE | APR_APPEND,
+                                 APR_OS_DEFAULT, pool));
 
   if (params.tunnel_user && run_mode != run_mode_tunnel)
     {
@@ -537,8 +707,14 @@ int main(int argc, const char *argv[])
           return svn_cmdline_handle_exit_error(err, pool, "svnserve: ");
         }
 
-      conn = svn_ra_svn_create_conn(NULL, in_file, out_file, pool);
-      svn_error_clear(serve(conn, &params, pool));
+      /* Use a subpool for the connection to ensure that if SASL is used
+       * the pool cleanup handlers that call sasl_dispose() (connection_pool)
+       * and sasl_done() (pool) are run in the right order. See issue #3664. */
+      connection_pool = svn_pool_create(pool);
+      conn = svn_ra_svn_create_conn2(NULL, in_file, out_file,
+                                     params.compression_level,
+                                     connection_pool);
+      svn_error_clear(serve(conn, &params, connection_pool));
       exit(0);
     }
 
@@ -609,10 +785,24 @@ int main(int argc, const char *argv[])
     {
       apr_socket_close(sock);
       family = APR_UNSPEC;
+
+      if (prefer_v6)
+        {
+          if (host == NULL)
+            host = "::";
+          sockaddr_info_flags = APR_IPV6_ADDR_OK;
+        }
+      else
+        {
+          if (host == NULL)
+            host = "0.0.0.0";
+          sockaddr_info_flags = APR_IPV4_ADDR_OK;
+        }
     }
 #endif
 
-  status = apr_sockaddr_info_get(&sa, host, family, port, 0, pool);
+  status = apr_sockaddr_info_get(&sa, host, family, port,
+                                 sockaddr_info_flags, pool);
   if (status)
     {
       err = svn_error_wrap_apr(status, _("Can't get address info"));
@@ -678,6 +868,34 @@ int main(int argc, const char *argv[])
     winservice_running();
 #endif
 
+  /* Configure FS caches for maximum efficiency with svnserve.
+   * For pre-forked (i.e. multi-processed) mode of operation,
+   * keep the per-process caches smaller than the default.
+   * Also, apply the respective command line parameters, if given. */
+  {
+    svn_cache_config_t settings = *svn_cache_config_get();
+
+    if (params.memory_cache_size != -1)
+      settings.cache_size = params.memory_cache_size;
+
+    settings.single_threaded = TRUE;
+    if (handling_mode == connection_mode_thread)
+      {
+#if APR_HAS_THREADS
+        settings.single_threaded = FALSE;
+#else
+        /* No requests will be processed at all
+         * (see "switch (handling_mode)" code further down).
+         * But if they were, some other synchronization code
+         * would need to take care of securing integrity of
+         * APR-based structures. That would include our caches.
+         */
+#endif
+      }
+
+    svn_cache_config_set(&settings);
+  }
+
   while (1)
     {
 #ifdef WIN32
@@ -685,9 +903,9 @@ int main(int argc, const char *argv[])
         return ERROR_SUCCESS;
 #endif
 
-      /* If we are using fulltext caches etc., we will allocate many large
-         chunks of memory of various sizes outside the cachde for those
-         fulltexts. Make sure, we use the memory wisely: use an allocator
+      /* If we are using fulltext caches etc. we will allocate many large
+         chunks of memory of various sizes outside the cache for those
+         fulltexts. Make sure we use the memory wisely: use an allocator
          that causes memory fragments to be given back to the OS early. */
 
       if (apr_allocator_create(&allocator))
@@ -697,7 +915,7 @@ int main(int argc, const char *argv[])
 
       /* Non-standard pool handling.  The main thread never blocks to join
          the connection threads so it cannot clean up after each one.  So
-         separate pools, that can be cleared at thread exit, are used */
+         separate pools that can be cleared at thread exit are used. */
 
       connection_pool = svn_pool_create_ex(NULL, allocator);
       apr_allocator_owner_set(allocator, connection_pool);
@@ -722,13 +940,29 @@ int main(int argc, const char *argv[])
           return svn_cmdline_handle_exit_error(err, pool, "svnserve: ");
         }
 
-      conn = svn_ra_svn_create_conn(usock, NULL, NULL, connection_pool);
+      /* Enable TCP keep-alives on the socket so we time out when
+       * the connection breaks due to network-layer problems.
+       * If the peer has dropped the connection due to a network partition
+       * or a crash, or if the peer no longer considers the connection
+       * valid because we are behind a NAT and our public IP has changed,
+       * it will respond to the keep-alive probe with a RST instead of an
+       * acknowledgment segment, which will cause svn to abort the session
+       * even while it is currently blocked waiting for data from the peer. */
+      status = apr_socket_opt_set(usock, APR_SO_KEEPALIVE, 1);
+      if (status)
+        {
+          /* It's not a fatal error if we cannot enable keep-alives. */
+        }
+
+      conn = svn_ra_svn_create_conn2(usock, NULL, NULL,
+                                     params.compression_level,
+                                     connection_pool);
 
       if (run_mode == run_mode_listen_once)
         {
           err = serve(conn, &params, connection_pool);
 
-          if (err && err->apr_err != SVN_ERR_RA_SVN_CONNECTION_CLOSED)
+          if (err)
             svn_handle_error2(err, stdout, FALSE, "svnserve: ");
           svn_error_clear(err);
 
@@ -745,7 +979,12 @@ int main(int argc, const char *argv[])
           if (status == APR_INCHILD)
             {
               apr_socket_close(sock);
-              svn_error_clear(serve(conn, &params, connection_pool));
+              err = serve(conn, &params, connection_pool);
+              log_error(err, params.log_file,
+                        svn_ra_svn_conn_remote_host(conn),
+                        NULL, NULL, /* user, repos */
+                        connection_pool);
+              svn_error_clear(err);
               apr_socket_close(usock);
               exit(0);
             }
@@ -755,7 +994,12 @@ int main(int argc, const char *argv[])
             }
           else
             {
-              /* Log an error, when we support logging. */
+              err = svn_error_wrap_apr(status, "apr_proc_fork");
+              log_error(err, params.log_file,
+                        svn_ra_svn_conn_remote_host(conn),
+                        NULL, NULL, /* user, repos */
+                        connection_pool);
+              svn_error_clear(err);
               apr_socket_close(usock);
             }
           svn_pool_destroy(connection_pool);
